/*
  This program transforms two DLV style programs to two first-order theories.
  If both are unsatisfiable, the programs are strongly equivalent.
  Copyright (C) 2004 Patrick Traxler

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <iostream>
#include <string>
#include <list>
#include <algorithm>
#include <sstream>
#include <stdio.h>
#include "st.h"

//////////////
//SymbolTable
////////////
SymbolTable st_CS, //Table of constant symbols.
            st_RS; //Table of relational symbols.

/*
    Adds s. s might have a arity, althought s is only
    identified by itself.
*/
void SymbolTable::add(const string& s, int arity) {
    if(s != "" && !exists(s))
        l.push_front(Symbol(s, arity));
}

/*
    Existence of s.
*/
bool SymbolTable::exists(const string& s) {
    list<Symbol>::iterator i;
    i = find(l.begin(), l.end(), Symbol(s));
    return i != l.end();
}

/*
    Returns the arity if s exists, -1 otherwise.
*/
int SymbolTable::getArity(const string& s) {
    list<Symbol>::iterator i;
    i = find(l.begin(), l.end(), Symbol(s));
    return i == l.end() ? -1 : (*i).arity;
}

/*
    Returns a symbol which does not occur in the 
    symbol table. 
    
    Note, that this function uses the fact that '_'
    does not occur in the used DLV syntax.
    
E.g.
First call:
    newSymbol("c")  returns "c_1"
    
Subsequent calls:
    newSymbol("c")  returns "c_2"
    newSymbol("sk") returns "sk_3"
    newSymbol("c")  returns "c_4"
*/
string SymbolTable::newSymbol(const string& prefix, int arity) {
    char t[sizeof(int)];
    sprintf(t, "%d", next_symbol);
    string s = prefix + "_" + t;
    next_symbol++;
    return s;
}

/*
    Iteration.
    
Usage: 
    SymbolTable st;
    ...
    string s;
    for (s = st.iterate(0); s != ""; st.iterate(1)) {
        ...
    }
*/
string SymbolTable::iterate(int pos) {
    if (pos == 0) { 
        if (l.empty()) 
            return "";

        itr = l.begin();
        return (*itr).s;
    }

    itr++;

    if (itr == l.end()) {
        return "";
    }

    return (*itr).s;
}

/*
    Printing.
*/
void SymbolTable::print() {
    list<Symbol>::const_iterator i;
    for (i = l.begin(); i != l.end(); i++) 
        cout << ((*i).s) << ' ';
    cout << endl;
}



////////
//Rule
/////

/*
    Replaces all occurences of t by u.
*/
void Rule::substitute(const string& t, const string& u) {
    int p = 0;    
    for (; (p = s.find(t, p)) != string::npos; p+= u.length())
        s.replace(p, t.length(), u);
}

/*
    Replaces "v" by ",", "not" by "~", and removes whitespaces.

    Note, that internally every rule is cleaned.
*/
void Rule::clean() {
    substitute(":-", " :- ");
    substitute("-", " - ");
    substitute(",", " , ");
    substitute("\t", " ");
    substitute("\n", " ");
    substitute(" v ", ",");
    substitute(" not ", "~");
    substitute(" ", "");
}

/*
    Returns true if there is a rule in DLV_prog, false otherwise.
    It also removes the first rule from DLV_prog and stores it.

Usage:
    Rule r;
    while (r.nextRule(DLV_prog)) {
        ...
    }
*/
bool Rule::nextRule(string& DLV_prog) {
    int pos = DLV_prog.find(".", 0);
    s = DLV_prog.substr(0, pos);
    clean();
    DLV_prog.erase(0, pos + 1);
    return s != "";
}

/*
    Skolemization.
    
E.g "r(X,Y):-s(X),t(Y)" to "r(sk_1,sk_2):-s(sk_1),t(sk_2)"
*/
void Rule::skolemize() {
    int j, i = 0;    
    do 
    {
        //Look for the next variable.
        for (; (s[i] < 'A' || s[i] > 'Z') && i < s.length(); i++);
        for (j = i; ((s[j] >= 'A' && s[j] <= 'Z') || 
                     (s[j] >= 'a' && s[j] <= 'z') ||
                     (s[j] >= '0' && s[j] <= '9')) && j < s.length(); j++);
       
        //Replace the varible by a new constant symbol.
        if (i != j) {
            substitute(s.substr(i, j-i), st_CS.newSymbol("sk"));
        }
    } while (i < s.length());    
}

/*
    Returns the next literal, ":-", or "".
    
E.g "r(X,Y):-s(X),t(Y)"
1. call: "r(X,Y)"
2. call: ":-"
3. call: "s(X)"
4. call: "t(Y)"
5. call: ""
*/
string Rule::nextLiteral() {
    if (s == "")
        return "";
        
    if (s[0] == ':' && s[1] == '-') {
        s.erase(0, 2);
        return ":-";
    }
    
    int  i;
    for (i = 0; i < s.length() && s[i] != ',' && s[i] != '(' && s[i] != ':'; i++);

    if (s[i] == '(') {
        for (; i < s.length() && s[i] != ')'; i++);    
        i++;
    }
        
    string t;    

    if (i == s.length()) {
        t = s;
        s = "";
        return t;
    }
    
    switch (s[i]) {
    case ',':
        t = s.substr(0, i);
        s.erase(0, i+1);
    break;
    case ':':
        t = s.substr(0, i);
        s.erase(0, i);
    break;
    }
    
    return t;
}
